""" Elasticsearch tasks
"""
import logging

from celery import shared_task

from core_elasticsearch_app.components.data import elasticsearch
from core_main_app.system import api as data_system_api

logger = logging.getLogger(__name__)


@shared_task
def index_all_data_from_template(template):
    """Index all data"""
    try:
        data = data_system_api.get_all_by_template(
            template=template, order_by_field=None
        )
        for document in data:
            try:
                elasticsearch.index_data(document)
            except Exception as e:
                logger.error(
                    f"ERROR : An error occurred while indexing data : {str(e)}"
                )
    except Exception as e:
        logger.error(f"ERROR : An error occurred while indexing data : {str(e)}")


@shared_task
def index_data(data_id):
    """Index a data"""
    try:
        data = data_system_api.get_data_by_id(data_id)
        try:
            elasticsearch.index_data(data)
        except Exception as e:
            logger.error(f"ERROR : An error occurred while indexing data : {str(e)}")
    except Exception as e:
        logger.error(f"ERROR : An error occurred while indexing data : {str(e)}")



""" Elasticsearch client for the CDCS
"""
from elasticsearch import Elasticsearch

from core_elasticsearch_app.settings import ELASTICSEARCH_HOST, ELASTICSEARCH_PORT


class ElasticsearchClient(object):
    """Elasticsearch client"""

    _es = None

    @classmethod
    def es(cls):
        """Return Elasticsearch connection
        Returns:
        """
        if cls._es is None:
            cls._es = Elasticsearch(
                [{"host": ELASTICSEARCH_HOST, "port": ELASTICSEARCH_PORT}]
            )
        return cls._es

    @classmethod
    def delete_index(cls, index_name):
        return cls.es().indices.delete(index=index_name, ignore=[400, 404])

    @classmethod
    def create_index(cls, index_name, settings=None):
        return cls.es().indices.create(index=index_name, body=settings, ignore=400)

    @classmethod
    def index_document(cls, index_name, document_id, document):
        return cls.es().index(index=index_name, id=document_id, body=document)

    @classmethod
    def delete_document(cls, index_name, document_id):
        return cls.es().delete(index=index_name, id=document_id)

    @classmethod
    def update_document(cls, index_name, document_id, document):
        return cls.es().update(index=index_name, id=document_id, body=document)

    @classmethod
    def search(cls, index_name, query):
        return cls.es().search(index=index_name, body=query)



""" MongoDB utils for the Data component
"""
import logging

from core_elasticsearch_app.utils.utils import get_nested_value
from core_main_app.utils.xml import xpath_to_dot_notation

logger = logging.getLogger(__name__)


def get_value_from_path(data, path):
    """Get value from xpath

    Args:
        data:
        path:

    Returns:

    """
    value = get_nested_value(data.dict_content, path)
    if value:
        if isinstance(value, dict):
            logger.info(
                "A dict was found at {0} for data {1}".format(path, str(data.id))
            )
            try:
                value = value["#text"]
            except:
                value = None
                logger.warning(
                    "#text was not found at path {0} for data {1}".format(
                        path, str(data.id)
                    )
                )
    else:
        logger.warning(
            "No value could be found at path {0} for data {1}".format(
                path, str(data.id)
            )
        )
    return value


def get_exists_query_from_path(path):
    """Return mongodb query that checks if path exists in database

    Args:
        path:

    Returns:

    """
    dot_path = xpath_to_dot_notation(path)
    return {
        "$or": [
            {f"dict_content.{dot_path}": {"$exists": True}},
            {f"dict_content.{dot_path}.#text": {"$exists": True}},
        ]
    }



""" Watchers for the data collection
"""

from core_elasticsearch_app.settings import ELASTICSEARCH_AUTO_INDEX
from core_elasticsearch_app.tasks import index_data
from core_main_app.components.data.models import Data
from signals_utils.signals.mongo import connector, signals


def post_save_data(sender, document, **kwargs):
    """Method executed after a saving of a Data object.
    Args:
        sender: Class.
        document: Data document.
        **kwargs: Args.

    """
    # only deal with data in a workspace
    if not document.workspace:
        return

    # only deal with public data
    from core_main_app.components.workspace import api as workspace_api

    public_workspaces = workspace_api.get_all_public_workspaces().values_list("id")
    if document.workspace.id not in public_workspaces:
        return

    # Creates or updates a document in the index.
    index_data.apply_async((str(document.id),))


def init():
    """Connect to Data object events."""
    if ELASTICSEARCH_AUTO_INDEX:
        connector.connect(post_save_data, signals.post_save, Data)



""" ElasticsearchTemplate model
"""
from django_mongoengine import fields, Document
from mongoengine import errors as mongoengine_errors
from mongoengine.queryset.base import CASCADE

from core_explore_keyword_app.components.search_operator.models import (
    validate_xpath_list,
)
from core_main_app.commons import exceptions
from core_main_app.components.template.models import Template


class ElasticsearchTemplate(Document):
    """ElasticsearchTemplate object"""

    template = fields.ReferenceField(
        Template, blank=False, reverse_delete_rule=CASCADE, unique=True
    )
    title_path = fields.StringField(default=None)
    description_paths = fields.ListField(
        blank=False, validation=validate_xpath_list, default=[]
    )

    @staticmethod
    def get_by_id(es_template_id):
        """Returns the object with a given id

        Args:
            es_template_id:

        Returns:
            ElasticsearchTemplate (obj): ElasticsearchTemplate

        """
        try:
            return ElasticsearchTemplate.objects.get(pk=es_template_id)
        except mongoengine_errors.DoesNotExist as e:
            raise exceptions.DoesNotExist(str(e))
        except Exception as e:
            raise exceptions.ModelError(str(e))

    @staticmethod
    def get_by_template(template):
        """Returns the object with a given template

        Args:
            template:

        Returns:
            ElasticsearchTemplate (obj): ElasticsearchTemplate

        """
        try:
            return ElasticsearchTemplate.objects.get(template=template)
        except mongoengine_errors.DoesNotExist as e:
            raise exceptions.DoesNotExist(str(e))
        except Exception as e:
            raise exceptions.ModelError(str(e))

    @staticmethod
    def get_all():
        """Returns all objects"""
        return ElasticsearchTemplate.objects.all()
